local slk = slk
local yi = yi
local setmetatable = setmetatable
local type = type

local function createObject(self, object)
    return setmetatable({}, {
        __index = function (t, tag)
			if 0 == tag then
				return self
			elseif 1 == tag then
				return object
			else
				local str = object:get(tag)
				if str then 
					return self._mgr:convert(str)
				else
					return '' 
				end
			end
        end
        })
end

local function createTable(self, table_type)
	return setmetatable({}, {
            __index = function (t, id)
				local cache = self._cache[table_type]
				if nil == cache then
					cache = self._mgr:load(table_type)
					self._cache[table_type] = cache
				end
				if 0 == id then
					return self
				elseif 1 == id then
					return cache
				else
					return createObject(self, cache:get(id) or slk.slk_object())
				end
            end
            })
end

local function createMiscTable(self)
	return setmetatable({}, {
            __index = function (t, id)
				local cache = self._cache[slk.OBJ_TYPE.MISC]
				if nil == cache then
					cache = self._mgr:load(slk.OBJ_TYPE.MISC)
					self._cache[slk.OBJ_TYPE.MISC] = cache
				end
				if 0 == id then
					return self
				elseif 1 == id then
					return cache:get('Misc')
				else
					local str = (cache:get('Misc') or slk.slk_object()):get(id)
					if str then 
						return str
					else
						return '' 
					end
				end
            end
            })
end

local SLK = {}

function SLK.Create(self)
	local mpq_path = yi.get_editor_root_dir() / "share" / "mpq"
	local obj = setmetatable({}, { __index = self})
	
	obj._mgr = slk.manager()
	obj._mgr:open((mpq_path / "units.mpq"):string())
	obj._mgr:open((mpq_path / "ydwe.mpq"):string())
	obj._mgr:attach(__map_handle__)
	obj._cache = {}		
	
	obj.Ability      = createTable(obj, slk.OBJ_TYPE.ABILITY)
	obj.Buff         = createTable(obj, slk.OBJ_TYPE.BUFF)
	obj.Destructable = createTable(obj, slk.OBJ_TYPE.DESTRUCTABLE)
	obj.Doodad       = createTable(obj, slk.OBJ_TYPE.DOODAD)
	obj.Item         = createTable(obj, slk.OBJ_TYPE.ITEM)
	obj.Unit         = createTable(obj, slk.OBJ_TYPE.UNIT)
	obj.Upgrade      = createTable(obj, slk.OBJ_TYPE.UPGRADE)
	obj.Misc         = createMiscTable(obj)
	
	return obj
end

function SLK.Foreach(Table)
	local It = Table[1]:begin()
	return function ()
		if not It:is_valid() then
			return nil
		end
		local k, v = It:key(), It:value()
		It:next()
		if type(v) == 'string' then
			return k, Table[0]._mgr:convert(v)
		else
			return k, createObject(Table[0], v)
		end
	end
end

return SLK
