-- 载入一些库
require "sys"
require "bit"
require "lfs"
require "std"
require "ex"
require "ydwelua"
require "titan_engine"
require "reg"

local stormlib = ar.stormlib

-- 本地魔兽版本号
local_warcraft_version = nil

-- YDWE的菜单句柄
ydwe_menu_handle = nil

-- 脚本用编码转换器
iconv_utf8_to_locale = nil

-- 主窗口句柄
main_window_handle = nil

-- 版本信息
yd_version = { major = 0, minor = 0, revision = 0, build = 0, full_string = "" }

-- 重新载入配置文件
function reload_configuration()
	global_config:reload(global_config_file)
end


-- 从主程序的mpq目录下载入MPQ
-- mpqname - MPQ的文件名
-- 返回值：MPQ句柄
local function load_mpq_in_mpq_dir(mpqname)
	local result = 0
	-- 获取主程序所在的目录
	local editor_dir = yi.get_editor_root_dir()
	-- 定位到主程序目录下mpq目录内的MPQ文件
	local mpq = editor_dir / "share" / "mpq" / mpqname

	-- 文件存在否？
	if fs.exists(mpq) then
		result = yi.native_open_archive(mpq, 14)
		if result then
			log.debug("Loaded " .. mpq:filename():string())
		else
			log.error("Cannot load " .. mpq:filename():string())
		end
	else
		log.error("Cannot find " .. mpq:filename():string())
	end

	return result
end

local function set_triggereditor_color()
	reg.SaveInt(reg.HKEY_CURRENT_USER, "Software\\Blizzard Entertainment\\WorldEdit\\Trigger Display Colors", "TC_YDHIDE",  0xffff0000)
	reg.SaveInt(reg.HKEY_CURRENT_USER, "Software\\Blizzard Entertainment\\WorldEdit\\Trigger Display Colors", "TC_COMMENT", 0xff008000)
end

-- 检查魔兽目录下是否有可能引起冲突的文件夹
local function check_conflicting_ui()
	log.debug("check_conflicting_ui")
	local war_path = yi.get_warcraft_dir()
	local file_list = {"ui/loading-yuexie.blp", "ui/miscdata.txt", "ui/triggerdata.txt", "ui/triggerstrings.txt", "ui/worldeditdata.txt", "ui/worldeditstrings.txt"}
	local found = false
	
	for index, file in ipairs(file_list) do
		if fs.exists(war_path / file) then
			found = true
			break
		end
	end
	
	if found then
		if sys.message_dialog(nil, string.format(_("YDWE has detected that there is a directory named \"%s\" located in Warcraft 3 installation directory. It may prevent YDWE from working. Do you want to delete it?"), 'UI'), _("YDWE"), bit.bor(ydwelua.MB_ICONQUESTION, ydwelua.MB_YESNO)) == ydwelua.IDYES then
			for index, file in ipairs(file_list) do			
				log.debug("remove file " .. (war_path / file):string())
				pcall(fs.remove_all, war_path / file)
			end
		end
	end
end

local function check_conflicting_units()
	log.debug("check_conflicting_units")
	local units_dir = yi.get_warcraft_dir() / 'Units'
	local found = false
	
	for file in lfs.dir(units_dir:string()) do	
		if not fs.is_directory(units_dir / file) then
			found = true
			break
		end
	end
	
	if found then
		if sys.message_dialog(nil, string.format(_("YDWE has detected that there is a directory named \"%s\" located in Warcraft 3 installation directory. It may prevent YDWE from working. Do you want to delete it?"), 'Units'), _("YDWE"), bit.bor(ydwelua.MB_ICONQUESTION, ydwelua.MB_YESNO)) == ydwelua.IDYES then		
			for file in lfs.dir(units_dir:string()) do	
				if not fs.is_directory(units_dir / file) then
					log.debug("remove file " .. (units_dir / file):string())
					pcall(fs.remove_all, units_dir / file)
				end
			end
		end
	end
end


-- 清除可能引起冲突的文件
local function clear_potential_conflicting()
	log.debug("clear_potential_conflicting")
	-- 获取魔兽路径
	local war_path = yi.get_warcraft_dir()
	-- 需要清理的文件列表
	local file_list = {"MiscData.txt", "TriggerData.txt", "TriggerStrings.txt", "WorldEditData.txt", "WorldEditLayout.txt", "WorldEditStrings.txt", "war3map.j", "blizzard.j", "Scripts/blizzard.j", "common.j", "Scripts/common.j"}

	for index, file in ipairs(file_list) do
		if fs.exists(war_path / file) then
			log.debug("remove file " .. (war_path / file):string())
			pcall(fs.remove, war_path / file)
		end
	end
end

	
-- 从魔兽中得到是否是1.24。通过搜索common.j有没有导出StringHash
-- 返回值：两个，魔兽版本，错误信息。如果没有错误，错误信息为nil
local function get_war3_version_from_script()
	local retval = local_warcraft_version
	local err = "Cannot extract file from warcraft"
	local common_j_path = yi.get_editor_root_dir() / "logs" / "common.j"
	local mpq_handle = stormlib.open_archive(yi.get_warcraft_dir() / 'War3Patch.mpq', 0, 0)
	if mpq_handle then
		if stormlib.has_file(mpq_handle, "common.j") then
			stormlib.extract_file(mpq_handle, common_j_path, "common.j")
		elseif stormlib.has_file(mpq_handle, "scripts\\common.j") then
			stormlib.extract_file(mpq_handle, common_j_path, "scripts\\common.j")
		else
			return retval, err
		end
		stormlib.close_archive(mpq_handle)
		
		local f, e = io.open(common_j_path:string(), "rb")
		if f then
			local s = f:read("*a")
			f:close()
			if s:find("StringHash") then
				retval = 1.24
			else
				retval = 1.20
			end
			err = nil
		else
			retval = local_warcraft_version
			err = e
		end
	end

	return retval, err
end

-- 检测魔兽的版本
local function check_war3_version()
	local version_info = sys.get_module_version_info(yi.get_warcraft_dir() / "game.dll")
	local_warcraft_version = tonumber(version_info.FileVersionMS)
	log.trace(string.format(
		(local_warcraft_version < 1.24 and "Old" or "New") ..
		" version warcraft %s detected", version_info.FileVersionMS
	))

	-- 检查“版本转换器”等造成的game.dll和war3patch.mpq不一致的问题
	if global_config:get_integer("MapSave.Option", 0) == 0 then
		-- 检测魔兽中包含的脚本文件所代表的版本
		local script_warcraft_version, e = get_war3_version_from_script()

		if e then
			log.warn("Cannot get warcraft 3 version from script: " .. e)
		else
			-- 二者如果不一致，则提示
			if (local_warcraft_version < 1.24) ~= (script_warcraft_version < 1.24) then
				sys.message_dialog(
					nil,
					_("YDWE has detected that your game.dll and war3patch.mpq mismatch. It may be caused by the so called \"Warcraft Version Converter\". This situation will cause a failure on saving and testing maps. It is strongly recommended to use the offical patched provided by Blizzard."),
					_("YDWE"),
					bit.bor(ydwelua.MB_ICONQUESTION, ydwelua.MB_OK)
				)
			end
		end
	end
end

-- WE的Hack函数，放这里避免污染全局名字空间
wehack = {}

-- BOOL __fastcall trigger_editor_editbox_copy(const char *source)
-- hook函数，执行转码
-- source - 目标字符串
-- 返回值：TRUE(1)表示成功，FALSE(0)表示失败
function wehack.trigger_editor_editbox_copy_hook(source)
	-- 刷新配置数据
	reload_configuration()

	-- 是否转编码？
	if global_config:get_integer("FeatureToggle.EnableTriggerCopyEncodingAutoConversion", 0) == 1 and source then
		-- 转换
		source = __(source)
	end

	-- 调用原来的函数完成工作
	return wehack.original_trigger_editor_editbox_copy(source)
end

-- 初始化Hack
-- 参数无，返回值无
local function initialize_we_hacks()
	-- CInvoke上下文，必须先初始化
	wehack.cinvoke_context = ccontext.new()

	-------------------------------------------------------------------------------------
	-- 自动转码

	-- 自动转码函数地址
	wehack.trigger_editor_editbox_copy = 0x71FE90
	-- 自动转码hook函数
	wehack.trigger_editor_editbox_copy_hook_callback
		= wehack.cinvoke_context:new_callback("fastcall", Clong, wehack.trigger_editor_editbox_copy_hook, Cstring)
	-- 利用TitanEngine新建一个hook
	local result = TE_HooksInsertNewRedirection(
		wehack.trigger_editor_editbox_copy,	-- hook目标地址
		mem.pointer_to_number(cinv.callback_to_ptr(wehack.trigger_editor_editbox_copy_hook_callback)), -- hook函数
		TEE_HOOK_NRM_JUMP	-- hook方式
	) and TE_HooksEnableRedirection(wehack.trigger_editor_editbox_copy, false)	-- false代表只启用一个hook而不是所有的
	if result then
		-- 通过trampoline构造原始函数
		wehack.original_trigger_editor_editbox_copy = wehack.cinvoke_context:create_function(
			"fastcall",
			Clong,
			TE_HooksGetTrampolineAddress(wehack.trigger_editor_editbox_copy),	-- 获得桥接地址
			Cstring
		)
		log.debug("Trigger editor copy hook installed successfully")
	else
		log.error("Trigger editor copy hook installing failed.")
	end
end


-- 显示制作者和感谢信息
function show_credit()
	yi.we_message_show("    ------------------------------------------------------------", 0)
	yi.we_message_show("           Welcome to YDWE, the best WE mod in China!           ", 0)
	yi.we_message_show("    ------------------------------------------------------------", 0)
	yi.we_message_show("    Director:   Everguo", 0)
	yi.we_message_show("    UI & Jass:  Fetrix_sai, Everguo, Warft_TigerCN, Wataloo & Xylign", 0)
	yi.we_message_show("    Program:    Actboy168, Aeris & Warft_TigerCN", 0)
	yi.we_message_show("    Art:        C kuhn", 0)
	yi.we_message_show("    Official Site: http://www.ydwe.net", 0)
	yi.we_message_show("    *** SPECIAL THANKS ***", 0)
	yi.we_message_show("    JassNewGenPack for ideas at www.wc3c.net", 0)
	yi.we_message_show("    Vexorian for his jasshelper compiler", 0)
	yi.we_message_show("    ADOLF and VD for their cjass compiler & TESH", 0)
	yi.we_message_show("    ...", 0)
	yi.we_message_show("    And all users & supporters, including", 0)
	yi.we_message_show("    YOU", 0)
end


-- 注入代码表
inject_code_table_new = {}
inject_code_table_old = {}

-- 扫描注入代码
-- config_dir - 需要扫描的路径
-- 返回值无，修改全局变量inject_code_table_new以及inject_code_table_old
-- inject_code_table_new - 新版（1.24）函数表
-- inject_code_table_old - 旧版函数表
local function scan_for_inject_code(config_dir)
	local counter = 0
	log.trace("Scanning for inject files in " .. config_dir:string())

	-- 遍历目录
	for file in lfs.dir(config_dir:string()) do
		local full_path = config_dir / file
		local full_path_string = full_path:string()
		local mode = lfs.attributes(full_path_string).mode

		-- 筛选合适的文件
		if mode == "file"		-- 是普通文件
			and #file > 4							-- 文件名长度大于4
			and file:sub(-4):lower() == ".cfg"		-- 扩展名为.cfg
		then
			-- 插入新表
			local table_new = {}
			local table_old = {}

			-- 解析状态，默认0
			-- 0 - 1.24/1.20通用
			-- 1 - 1.24专用
			-- 2 - 1.20专用
			local state = 0

			-- 循环处理每一行
			for line in io.lines(full_path_string) do
				-- 插入函数名
				local trimed = line:trim()
				if trimed ~= "" and trimed:sub(1, 1) ~= "#" then
					if trimed == "[general]" then
						state = 0
					elseif trimed == "[new]" then
						state = 1
					elseif trimed == "[old]" then
						state = 2
					else
						if state == 0 then
							table.insert(table_new, trimed)
							table.insert(table_old, trimed)
						elseif state == 1 then
							table.insert(table_new, trimed)
						elseif state == 2 then
							table.insert(table_old, trimed)
						end
					end
				end
			end

			-- 插入全局表中（替换文件扩展名）
			local substitution = config_dir / (file:sub(1, -5) .. ".j")
			if #table_old > 0 then
				inject_code_table_old[substitution] = table_old
			end

			if #table_new > 0 then
				inject_code_table_new[substitution] = table_new
			end

			counter = counter + 1
			--log.trace("Scanned " .. file)
		elseif mode == "directory"				-- 是目录
			and file ~= "." and file ~= ".."	-- 不是特殊目录
		then
			-- 递归处理
			counter = counter + scan_for_inject_code(full_path)
		end
	end
	return counter
end

plugin = {}
plugin.loaders = {}

plugin.load = function (plugin_config_path)
	log.trace("Load plugin config " .. plugin_config_path:string())
	
	local plugin_config = sys.config_property(plugin_config_path)
	local plugin_name = plugin_config:get_string('Info.PluginName', '')
	
	if plugin_name == '' then
		log.error("Cannot plugin name.")
		return
	end
	
	if plugin.loaders[plugin_name] then
		log.error(plugin_name .. " already exists.")
		return
	end
	
	if 0 == plugin_config:get_integer('Load.Enable', 0) then
		log.debug("Diable " .. plugin_name .. ".")
		return		
	end
	
	local plugin_loader_path = plugin_config:get_string('Load.Loader', '')
	if plugin_loader_path == '' then
		log.error("Cannot find " .. plugin_name .. "'s loader.")
		return 
	end
	plugin_loader_path = plugin_config_path:parent_path() / plugin_loader_path
	if not fs.exists(plugin_loader_path) then
		log.error(plugin_name .. "'loader does not exist.")
		return
	end
	
	local s, r = pcall(dofile, plugin_loader_path:string())
	if not s then
		log.error("Error in initialize " .. plugin_name .. "'s loader: ".. r)
		return 
	end
	
	plugin.loaders[plugin_name] = r

	local plugin_dll_path = plugin_config:get_string('Load.Dll', '')
	if plugin_dll_path == '' then
		log.error("Cannot find " .. plugin_name .. "'s dll.")
		return 
	end
	plugin_dll_path = plugin_config_path:parent_path() / plugin_dll_path
	if not fs.exists(plugin_dll_path) then
		log.error(plugin_name .. "'dll does not exist. " .. plugin_dll_path:string())
		return
	end
	
	s, r, msg = pcall(plugin.loaders[plugin_name].load, plugin_dll_path)
		
	if not s then
		log.error("Error in load " .. plugin_name .. ": ".. r)
		plugin.loaders[plugin_name] = nil
		return 
	end
	
	if not r then
		if msg then
			log.error(plugin_name .. " loading failed: " .. msg ..".")
		else
			log.error(plugin_name .. " loading failed.")
		end
		plugin.loaders[plugin_name] = nil
		return
	end
	
	log.debug(plugin_name .. " loaded successfully.")	
	return
end

plugin.scan = function (plugin_dir)
	-- 遍历目录
	for file in lfs.dir(plugin_dir:string()) do
		local full_path = plugin_dir / file
		local full_path_string = full_path:string()
		local mode = lfs.attributes(full_path_string).mode

		-- 筛选合适的文件
		if mode == "file"		-- 是普通文件
			and #file > 6							    -- 文件名长度大于6
			and file:sub(-6):lower() == ".plcfg"		-- 扩展名为.plcfg
		then
			plugin.load(full_path)
		elseif mode == "directory"				-- 是目录
			and file ~= "." and file ~= ".."	-- 不是特殊目录
		then
			plugin.scan(full_path)
		end
	end
end

plugin.unload_all = function ()
	for name, loader in pairs(plugin.loaders) do
		log.trace("Unload plugin " .. name .. ".")
		pcall(loader.unload)
	end
end

-- 本函数在编辑器启动时调用，可以在本函数中载入一些插件
-- event_data - 事件参数
-- 	暂无内容
-- 返回值：返回非负数表示成功，负数表示失败
function on_startup(event_data)
	log.debug("********************* on startup start *********************")

	-- 检测YDWE自己的版本
	local version_info = sys.get_module_version_info(yi.get_editor_root_dir() / "bin" / "YDWEInject.dll")
	yd_version.major = tonumber(version_info.FileVersionMSH)
	yd_version.minor = tonumber(version_info.FileVersionMSL)
	yd_version.revision = tonumber(version_info.FileVersionLSH)
	yd_version.build = tonumber(version_info.FileVersionLSL)
	yd_version.full_string = version_info.FileVersion

	-- 刷新配置数据
	reload_configuration()

	-- 检测UI和Units目录
	check_conflicting_ui()
	check_conflicting_units()

	-- 检查魔兽目录下是否有可能引起冲突的文件夹
	clear_potential_conflicting()

	-- 检测魔兽的版本
	check_war3_version()	

	-- 载入Patch MPQ
	load_mpq_in_mpq_dir("units.mpq")
	load_mpq_in_mpq_dir("ydwe.mpq")
	
	-- 加载插件
	plugin.scan(yi.get_editor_root_dir() / "plugin")

	-- 载入注入代码配置
	local counter = scan_for_inject_code(yi.get_editor_root_dir() / "jass")
	log.trace(string.format("Scanned file: %d", counter))

	-- 安装其他的hook、patch等
	initialize_we_hacks()
		
	set_triggereditor_color()

	-- 显示感谢信息
	show_credit()

	log.debug("********************* on startup end *********************")

	return 0
end

-- 本函数在编辑器关闭时调用，可以在本函数中作一些清理工作
-- event_data - 事件参数。table类型，暂时没有内容。
-- 返回值：一定要返回0
function on_exit(event_data)
	log.debug("********************* on exit start *********************")

	plugin.unload_all()

	log.debug("********************* on exit end *********************")

	return 0
end

-- 本函数在测试地图时使用
-- event_data - 事件参数，table，包含以下值
--	map_path - 保存的地图路径，字符串类型
--	application_name - 应用程序名，字符串对象
--	command_line - 命令行，字符串对象
-- 返回：非负表示成功，负数表示失败。如成功且无特殊情况原则上应该返回0
function on_test(event_data)
	log.debug("********************* on test start *********************")

	-- 刷新配置数据
	reload_configuration()
	
	-- 获取当前测试的地图名
	local map_path = fs.path(event_data.map_path)
	log.debug("Testing " .. map_path:string())
	log.debug("Testing " .. event_data.command_line)

	-- 附加命令行
	local command_aux = ""

	-- 是否OpenGL方式？
	if global_config:get_integer("MapTest.LaunchOpenGL", 0) == 1 then
		command_aux = command_aux .. " -opengl"
	end

	-- 是否窗口方式？
	if global_config:get_integer("MapTest.LaunchWindowed", 1) == 1 then
		command_aux = command_aux .. " -window"
	end
	
	command_aux = command_aux .. ' -ydwe "' .. yi.get_editor_root_dir():string() .. '"'

	local new_command_line
	local b, e = event_data.command_line:find(' %-loadfile ')
	if b then
		new_command_line = event_data.command_line:sub(1, b-1) .. ' -loadfile "'.. event_data.command_line:sub(e+1, -1) .. '"' .. command_aux
	else
		new_command_line = event_data.command_line .. command_aux
	end
	log.debug("Testing " .. new_command_line)
	
	local result = false
	-- 启动魔兽开始测试...
	local war3_helper_dll = yi.get_editor_root_dir() / "plugin" / "warcraft3" / "yd_loader.dll"
	if fs.exists(war3_helper_dll) then
		result = sys.process_spawn_with_dll(event_data.application_name, new_command_line, nil, war3_helper_dll:string(), false)
	else
		result = sys.process_spawn(event_data.application_name, new_command_line, nil, false)
	end

	log.debug("********************* on test end *********************")
	if result then return 0 else return -1 end
end

-- 在WE载入MSS引擎解码器时调用，过滤载入的dll
-- event_data - 事件参数，table，包含以下值
--	library_name - 解码器路径，字符串
-- 返回非负数表示允许载入，负数表示不允许。无特殊情况一般返回0
function on_mss_load_provider_library(event_data)
	log.debug("Loading provider " .. event_data.library_name)

	-- 全部放行
	return 0
end

----------------------------------------------------------------------

-- 转换为本地语言
-- str - UTF-8编码的文本ID
-- 返回：本地语言、本地编码（操作系统相关）的字符串
-- 注意：1. 找不到翻译信息的则保留英文
--       2. 本文件编码为UTF-8，因此所有硬编码的字符串均为UTF-8编码，要注意转换
function _(str)
	return iconv_utf8_to_locale:iconv(i18n.gettext(str))
end

-- 转换为本地编码
-- str - UTF-8编码的文本
-- 返回：本地编码（操作系统相关）的字符串
-- 注意：本文件编码为UTF-8，因此所有硬编码的字符串均为UTF-8编码，要注意转换
function __(str)
	return iconv_utf8_to_locale:iconv(str)
end

-- 扩展数学库
mathx = {}

-- 比较2个浮点数是否相等
-- a, b - 要比较的2个浮点数
-- eps - 精确度阈值
-- 返回值：true表示2个数的差在精确阈值内（相等），false表明不在阈值内
function mathx.feq(a, b, eps)
	if not eps then eps = 1e-10 end
	return math.abs(a - b) < eps
end

-- 初始化编码转换器
local function initialize_encode_converter()
	iconv_utf8_to_locale = iconv.open(iconv.locale_charset() .. "//TRANSLIT//IGNORE", "UTF-8")
	iconv_locale_to_utf8 = iconv.open("UTF-8//TRANSLIT//IGNORE", iconv.locale_charset())
end

-- 初始化国际化组件
local function initialize_gettext()
	i18n.setlocale(1729, "")	-- 1729 - LC_MESSAGES
	i18n.textdomain("MainScript");
	i18n.bindtextdomain("MainScript", (yi.get_editor_root_dir() / "share" / "locale"):string())
end


-- 初始化代码
local function initialize()
	-- 初始化编码转换器
	initialize_encode_converter()

	-- 初始化国际化组件
	initialize_gettext()

	-- 初始化配置文件
	global_config_file = yi.get_editor_root_dir() / "bin" / "EverConfig.cfg"
	global_config = sys.config_property(global_config_file)
end

-- 初始化
initialize()
